package meerkat.bulletinboard.sqlserver;

import java.security.InvalidKeyException;
import java.security.cert.CertificateException;
import java.sql.*;
import java.util.*;

import com.google.protobuf.ByteString;
import com.google.protobuf.ProtocolStringList;

import meerkat.bulletinboard.BulletinBoardServer;
import meerkat.bulletinboard.sqlserver.mappers.*;

import meerkat.comm.CommunicationException;

import meerkat.crypto.BatchDigest;
import meerkat.crypto.DigitalSignature;
import meerkat.crypto.concrete.ECDSASignature;
import meerkat.crypto.concrete.SHA256Digest;

import meerkat.protobuf.BulletinBoardAPI.*;
import meerkat.protobuf.Crypto.Signature;
import meerkat.protobuf.Crypto.SignatureVerificationKey;


import static meerkat.bulletinboard.sqlserver.BulletinBoardSQLServer.SQLQueryProvider.*;

import javax.sql.DataSource;

import org.springframework.jdbc.core.namedparam.MapSqlParameterSource;
import org.springframework.jdbc.core.namedparam.NamedParameterJdbcTemplate;
import org.springframework.jdbc.support.GeneratedKeyHolder;
import org.springframework.jdbc.support.KeyHolder;



/**
 * This is a generic SQL implementation of the BulletinBoardServer API.
 */
public class BulletinBoardSQLServer implements BulletinBoardServer{

	/**
	 * This interface provides the required implementation-specific data to enable an access to an actual SQL server.
	 * It accounts for differences in languages between SQL DB types and for the different addresses needed to access them.
	 */
	public interface SQLQueryProvider {

		/**
		 * Allowed query types.
		 * Note that each query returned has to comply with the parameter names specified ny getParamNames
		 */
		public static enum QueryType {

			FIND_MSG_ID(new String[] {"MsgId"}),
			INSERT_MSG(new String[] {"MsgId","Msg"}),
			INSERT_NEW_TAG(new String[] {"Tag"}),
			CONNECT_TAG(new String[] {"EntryNum","Tag"}),
			ADD_SIGNATURE(new String[] {"EntryNum","SignerId","Signature"}),
			GET_SIGNATURES(new String[] {"EntryNum"}),
			GET_MESSAGES(new String[] {}),
			GET_BATCH_MESSAGE_ENTRY(new String[] {"SignerId", "BatchId"}),
			CHECK_BATCH_LENGTH(new String[] {"SignerId", "BatchId"}),
			GET_BATCH_MESSAGE_DATA(new String[] {"SignerId", "BatchId", "StartPosition"}),
			INSERT_BATCH_DATA(new String[] {"SignerId", "BatchId", "SerialNum", "Data"}),
			CONNECT_BATCH_TAG(new String[] {"SignerId", "BatchId", "Tag"});

			private String[] paramNames;

			private QueryType(String[] paramNames) {
				this.paramNames = paramNames;
			}

			public String[] getParamNames() {
				return paramNames;
			}

			public String getParamName(int num) {
				return paramNames[num];
			}

		}

		/**
		 * This enum provides the standard translation between a filter type and the corresponding parameter name in the SQL query
		 */
		public static enum FilterTypeParam {

			ENTRY_NUM("EntryNum", Types.INTEGER),
			MSG_ID("MsgId", Types.BLOB),
			SIGNER_ID("SignerId", Types.BLOB),
			TAG("Tag", Types.VARCHAR),
			LIMIT("Limit", Types.INTEGER);

			private FilterTypeParam(String paramName, int paramType) {
				this.paramName = paramName;
				this.paramType = paramType;
			}

			private String paramName;
			private int paramType;

			public static FilterTypeParam getFilterTypeParamName(FilterType filterType) {
				switch (filterType) {

					case MSG_ID:
						return MSG_ID;

					case EXACT_ENTRY: // Go through
					case MAX_ENTRY:
						return ENTRY_NUM;

					case SIGNER_ID:
						return SIGNER_ID;

					case TAG:
						return TAG;

					case MAX_MESSAGES:
						return LIMIT;

					default:
						return null;
				}
			}

			public String getParamName() {
				return paramName;
			}

			public int getParamType() {
				return paramType;
			}
		}

		/**
		 * This function translates a QueryType into an actual SQL query.
		 * @param queryType is the type of query requested
         * @return a string representation of the query for the specific type of SQL database implemented.
         */
		public String getSQLString(QueryType queryType) throws IllegalArgumentException;

		/**
		 * Used to retrieve a condition to add to an SQL statement that will make the result comply with the filter type
		 * @param filterType is the filter type
		 * @param serialNum is a unique number used to identify the condition variables from other condition instances
		 * @return The SQL string for the condition
		 * @throws IllegalArgumentException if the filter type used is not supported
		 */
		public String getCondition(FilterType filterType, int serialNum) throws IllegalArgumentException;

		public String getConditionParamTypeName(FilterType filterType) throws IllegalArgumentException;

		/**
		 * @return the string needed in order to connect to the DB.
         */
		public DataSource getDataSource();

		/**
		 * This is used to get a list of queries that together create the schema needed for the DB.
		 * Note that these queries should not assume anything about the current state of the DB.
		 * In particular: they should not erase any existing tables and/or entries.
		 * @return The list of queries.
         */
		public List<String> getSchemaCreationCommands();

		/**
		 * This is used to get a list of queries that together delete the schema needed for the DB.
		 * This is useful primarily for tests, in which we want to make sure we start with a clean DB.
		 * @return The list of queries.
		 */
		public List<String> getSchemaDeletionCommands();

	}

	/**
	 * This method returns the value of the parameter specified in a message filter
	 * @param messageFilter is the filter
	 * @return the object parameter for the SQL query embedded in the filter (this depends on the filter type)
     */
	private Object getParam(MessageFilter messageFilter) {

		switch (messageFilter.getType()) {

			case MSG_ID: // Go through
			case SIGNER_ID:
				return messageFilter.getId().toByteArray();

			case EXACT_ENTRY: // Go through
			case MAX_ENTRY:
				return messageFilter.getEntry();

			case TAG:
				return messageFilter.getTag();

			case MAX_MESSAGES:
				return messageFilter.getMaxMessages();

			default:	// Unsupported filter type
				return null;
		}

	}

	/**
	 * This class implements a comparator for the MessageFilter class
	 * The comparison is done solely by comparing the type of the filter
	 * This is used to sort the filters by type
	 */
	public class FilterTypeComparator implements Comparator<MessageFilter> {

		@Override
		public int compare(MessageFilter filter1, MessageFilter filter2) {
			return filter1.getTypeValue() - filter2.getTypeValue();
		}
	}

	protected SQLQueryProvider sqlQueryProvider;

	protected NamedParameterJdbcTemplate jdbcTemplate;
	
	protected BatchDigest digest;
	protected DigitalSignature signer;
	
	protected List<SignatureVerificationKey> trusteeSignatureVerificationArray;
	protected int minTrusteeSignatures;

	protected List<List<SignatureVerificationKey>> pollingCommitteeSignatureVerificationKeyArray;
	protected int minCommiteeSignatures;

	/**
	 * This constructor sets the type of SQL language in use.
	 * @param sqlQueryProvider is the provider of the SQL query strings required for actual operation of the server.
     */
	public BulletinBoardSQLServer(SQLQueryProvider sqlQueryProvider) {
		this.sqlQueryProvider = sqlQueryProvider;
	}

	/**
	 * This method creates the schema in the given DB to prepare for future transactions
	 * It does not assume anything about the current state of the database
	 * @throws SQLException
     */
	private void createSchema() throws SQLException {

		final int TIMEOUT = 20;

		for (String command : sqlQueryProvider.getSchemaCreationCommands()) {
			jdbcTemplate.update(command,(Map) null);
		}

	}

	/**
	 * This method initializes the signatures, connects to the DB and creates the schema (if required).
	 */
	@Override
	public void init(String meerkatDB) throws CommunicationException {
		// TODO write signature reading part.
		
		digest = new SHA256Digest();
		signer = new ECDSASignature();

		jdbcTemplate = new NamedParameterJdbcTemplate(sqlQueryProvider.getDataSource());

		try {
			createSchema();
		} catch (SQLException e) {
			throw new CommunicationException("Couldn't create schema " + e.getMessage());
		}

	}

	/**
	 * This method verifies the authenticity of the received message based on
	 * the stored signatures.
	 * 
	 * @param msg
	 *            is the message to authenticate (containing the signature).
	 * @return TRUE if the message is authenticated and FALSE otherwise.
	 */
	private boolean verifyMessage(BulletinBoardMessage msg) {
		//TODO: Replace with actual verification.
		return true;
	}
	
	/**
	 * This procedure makes sure that all tags in the given list have an entry in the tags table. 
	 * @param tags
	 */
	protected void insertNewTags(String[] tags) throws SQLException {

		String sql;

		sql = sqlQueryProvider.getSQLString(QueryType.INSERT_NEW_TAG);
		Map namedParameters[] = new HashMap[tags.length];

		for (int i = 0 ; i < tags.length ; i++){
			namedParameters[i] = new HashMap();
			namedParameters[i].put(QueryType.INSERT_NEW_TAG.getParamName(0), tags[i]);
		}

		jdbcTemplate.batchUpdate(sql, namedParameters);

	}
	
	/**
	 * This procedure is used to convert a boolean to a BoolMsg.
	 * @param b is the boolean to convert.
	 * @return a ProtoBuf message with boolean payload.
	 */
	private BoolMsg boolToBoolMsg(boolean b){
		return BoolMsg.newBuilder()
			.setValue(b)
			.build();
	}

	@Override
	public BoolMsg postMessage(BulletinBoardMessage msg) throws CommunicationException {
		
		if (!verifyMessage(msg)) {
			return boolToBoolMsg(false);
		}

		String sql;
		Map[] namedParameterArray;

		byte[] msgID;
		long entryNum;
		
		ProtocolStringList tagList;
		String[] tags;
		
		List<Signature> signatureList;
		Signature[] signatures;
		
		// Calculate message ID (depending only on the the unsigned message)
		
		digest.reset();
		digest.update(msg.getMsg());
		
		msgID = digest.digest();
		
		// Add message to table if needed and store entry number of message.

			
		sql = sqlQueryProvider.getSQLString(QueryType.FIND_MSG_ID);
		Map namedParameters = new HashMap();
		namedParameters.put(QueryType.FIND_MSG_ID.getParamName(0),msgID);

		List<Long> entryNums = jdbcTemplate.query(sql, new MapSqlParameterSource(namedParameters), new LongMapper());

		if (entryNums.size() > 0){

			entryNum = entryNums.get(0);

		} else{

			sql = sqlQueryProvider.getSQLString(QueryType.INSERT_MSG);
			namedParameters.put(QueryType.INSERT_MSG.getParamName(0), msg.getMsg().toByteArray());

			KeyHolder keyHolder = new GeneratedKeyHolder();
			jdbcTemplate.update(sql,new MapSqlParameterSource(namedParameters),keyHolder);

			entryNum = keyHolder.getKey().longValue();

		}
		
		// Retrieve tags and store new ones in tag table.
		
		try {
			
		    tagList = msg.getMsg().getTagList();
		    tags = new String[tagList.size()];
		    tags = tagList.toArray(tags);
		    
		    insertNewTags(tags);
		    
		} catch (SQLException e) {
			throw new CommunicationException(e.getMessage());
		}
		
		// Connect message to tags.

		sql = sqlQueryProvider.getSQLString(QueryType.CONNECT_TAG);

		namedParameterArray = new HashMap[tags.length];

		for (int i = 0 ; i < tags.length ; i++) {
			namedParameterArray[i] = new HashMap();
			namedParameterArray[i].put(QueryType.CONNECT_TAG.getParamName(0), entryNum);
			namedParameterArray[i].put(QueryType.CONNECT_TAG.getParamName(1), tags[i]);
		}

		jdbcTemplate.batchUpdate(sql, namedParameterArray);
		
		// Retrieve signatures.
		
	    signatureList = msg.getSigList();
	    signatures = new Signature[signatureList.size()];
	    signatures = signatureList.toArray(signatures);
		
		// Connect message to signatures.

		sql = sqlQueryProvider.getSQLString(QueryType.ADD_SIGNATURE);

		namedParameterArray = new HashMap[signatures.length];

		for (int i = 0 ; i < signatures.length ; i++) {
			namedParameterArray[i] = new HashMap();
			namedParameterArray[i].put(QueryType.ADD_SIGNATURE.getParamName(0), entryNum);
			namedParameterArray[i].put(QueryType.ADD_SIGNATURE.getParamName(1), signatures[i].getSignerId().toByteArray());
			namedParameterArray[i].put(QueryType.ADD_SIGNATURE.getParamName(2), signatures[i].toByteArray());
		}

		jdbcTemplate.batchUpdate(sql,namedParameterArray);

		return boolToBoolMsg(true);
	}

	@Override
	public BulletinBoardMessageList readMessages(MessageFilterList filterList) throws CommunicationException {

		BulletinBoardMessageList.Builder resultListBuilder = BulletinBoardMessageList.newBuilder();

		// SQL length is roughly 50 characters per filter + 50 for the query itself
		StringBuilder sqlBuilder = new StringBuilder(50 * (filterList.getFilterCount() + 1));

		MapSqlParameterSource namedParameters;
		int paramNum;

		MessageMapper messageMapper = new MessageMapper();
		SignatureMapper signatureMapper = new SignatureMapper();

		List<MessageFilter> filters = new ArrayList<MessageFilter>(filterList.getFilterList());

		boolean isFirstFilter = true;

		Collections.sort(filters, new FilterTypeComparator());

		// Check if Tag/Signature tables are required for filtering purposes

		sqlBuilder.append(sqlQueryProvider.getSQLString(QueryType.GET_MESSAGES));
		// Add conditions

		namedParameters = new MapSqlParameterSource();

		if (!filters.isEmpty()) {
			sqlBuilder.append(" WHERE ");

			for (paramNum = 0 ; paramNum < filters.size() ; paramNum++) {

				MessageFilter filter = filters.get(paramNum);

				if (filter.getType().getNumber() != FilterType.MAX_MESSAGES_VALUE) {
					if (isFirstFilter) {
						isFirstFilter = false;
					} else {
						sqlBuilder.append(" AND ");
					}
				}

				sqlBuilder.append(sqlQueryProvider.getCondition(filter.getType(), paramNum));

				FilterTypeParam filterTypeParam = FilterTypeParam.getFilterTypeParamName(filter.getType());

				namedParameters.addValue(
						filterTypeParam.getParamName() + Integer.toString(paramNum),
						getParam(filter),
						filterTypeParam.getParamType(),
						sqlQueryProvider.getConditionParamTypeName(filter.getType()));

			}

		}

		// Run query

		List<BulletinBoardMessage.Builder> msgBuilders = jdbcTemplate.query(sqlBuilder.toString(), namedParameters, messageMapper);

		// Compile list of messages

		for (BulletinBoardMessage.Builder msgBuilder : msgBuilders) {

			// Retrieve signatures

			namedParameters = new MapSqlParameterSource();
			namedParameters.addValue(QueryType.GET_SIGNATURES.getParamName(0), msgBuilder.getEntryNum());

			List<Signature> signatures = jdbcTemplate.query(
					sqlQueryProvider.getSQLString(QueryType.GET_SIGNATURES),
					namedParameters,
					signatureMapper);

			// Append signatures
			msgBuilder.addAllSig(signatures);

			// Finalize message and add to message list.

			resultListBuilder.addMessage(msgBuilder.build());

		}

		//Combine results and return.
		return resultListBuilder.build();

	}

	/**
	 * This method checks if a specified batch exists and is already closed
	 * @param signerId is the ID of the publisher of the batch
	 * @param batchId is the unique (per signer) batch ID
     * @return TRUE if the batch is closed and FALSE if it is still open or doesn't exist at all
     */
	private boolean isBatchClosed(ByteString signerId, int batchId){

		String sql = sqlQueryProvider.getSQLString(QueryType.GET_BATCH_MESSAGE_ENTRY);

		MapSqlParameterSource namedParameters = new MapSqlParameterSource();
		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_ENTRY.getParamName(0), signerId);
		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_ENTRY.getParamName(1), batchId);

		List<Long> result = jdbcTemplate.query(sql,namedParameters,new LongMapper());

		return (result.size() > 0);

	}

	@Override
	public BoolMsg beginBatch(BeginBatchMessage message) throws CommunicationException {

		// Check if batch is closed
		if (isBatchClosed(message.getSignerId(), message.getBatchId())) {
			return BoolMsg.newBuilder().setValue(false).build();
		}

		// Add new tags to table
		ProtocolStringList tagList = message.getTagList();
		String[] tags = new String[tagList.size()];
		tags = tagList.toArray(tags);
		try {
			insertNewTags(tags);
		} catch (SQLException e) {
			throw new CommunicationException(e.getMessage());
		}

		// Connect tags
		String sql = sqlQueryProvider.getSQLString(QueryType.CONNECT_BATCH_TAG);
		MapSqlParameterSource namedParameters[] = new MapSqlParameterSource[tags.length];

		for (int i=0 ; i < tags.length ; i++) {
			namedParameters[i] = new MapSqlParameterSource();
			namedParameters[i].addValue(QueryType.CONNECT_BATCH_TAG.getParamName(0),message.getSignerId());
			namedParameters[i].addValue(QueryType.CONNECT_BATCH_TAG.getParamName(1),message.getBatchId());
			namedParameters[i].addValue(QueryType.CONNECT_BATCH_TAG.getParamName(2),tags[i]);
		}

		jdbcTemplate.batchUpdate(sql,namedParameters);

		return BoolMsg.newBuilder().setValue(true).build();
	}

	@Override
	public BoolMsg postBatchMessage(BatchMessage batchMessage) {

		// Check if batch is closed
		if (isBatchClosed(batchMessage.getSignerId(), batchMessage.getBatchId())) {
			return BoolMsg.newBuilder().setValue(false).build();
		}

		// Add data
		String sql = sqlQueryProvider.getSQLString(QueryType.INSERT_BATCH_DATA);
		MapSqlParameterSource namedParameters = new MapSqlParameterSource();

		namedParameters.addValue(QueryType.INSERT_BATCH_DATA.getParamName(0),batchMessage.getSignerId());
		namedParameters.addValue(QueryType.INSERT_BATCH_DATA.getParamName(1),batchMessage.getBatchId());
		namedParameters.addValue(QueryType.INSERT_BATCH_DATA.getParamName(2),batchMessage.getSerialNum());
		namedParameters.addValue(QueryType.INSERT_BATCH_DATA.getParamName(3),batchMessage.getData());

		jdbcTemplate.update(sql, namedParameters);

		return BoolMsg.newBuilder().setValue(true).build();

	}

	@Override
	public BoolMsg closeBatchMessage(CloseBatchMessage message) throws CommunicationException {

		// Check batch size

		String sql = sqlQueryProvider.getSQLString(QueryType.CHECK_BATCH_LENGTH);
		MapSqlParameterSource namedParameters = new MapSqlParameterSource();

		namedParameters.addValue(QueryType.CHECK_BATCH_LENGTH.getParamName(0),message.getSig().getSignerId());
		namedParameters.addValue(QueryType.CHECK_BATCH_LENGTH.getParamName(1),message.getBatchId());

		List<Long> lengthResult = jdbcTemplate.query(sql, namedParameters, new LongMapper());

		if (lengthResult.get(0) != message.getBatchLength()) {
			return BoolMsg.newBuilder().setValue(false).build();
		}

		// Check signature

		sql = sqlQueryProvider.getSQLString(QueryType.GET_BATCH_MESSAGE_DATA);
		namedParameters = new MapSqlParameterSource();

		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(0),message.getSig().getSignerId());
		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(1),message.getBatchId());
		namedParameters.addValue(QueryType.GET_BATCH_MESSAGE_DATA.getParamName(2),0); // Read from the beginning

		List<BatchData> batchDataList = jdbcTemplate.query(sql, namedParameters, new BatchDataMapper());

		try {
			signer.initVerify(message.getSig());
		} catch (CertificateException | InvalidKeyException e) {
			return BoolMsg.newBuilder().setValue(false).build();
		}

		//TODO: FInish this

		//signer.updateContent(msgStream);
		//assertTrue("Signature did not verify!", signer.verify());

		return null;
	}

	@Override
	public List<BatchData> readBatch(BatchSpecificationMessage message) {
		return null; // TODO: Implement this
	}

	@Override
	public void close() {}

}
